"""Agent provider implementations."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, Protocol, runtime_checkable

from psygnal import Signal
from pydantic_ai.result import FinalResult
from pydantic_graph import BaseNode, End
from typing_extensions import TypeVar

from llmling_agent.log import get_logger
from llmling_agent.tools import ToolCallInfo


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, AsyncIterator
    from contextlib import AbstractAsyncContextManager

    from pydantic_ai.agent import AgentRunResult
    import tokonomics
    from tokonomics.pydanticai_cost import Usage

    from llmling_agent.agent.context import AgentContext
    from llmling_agent.common_types import ModelProtocol, ModelType
    from llmling_agent.messaging.messages import ChatMessage, TokenCost
    from llmling_agent.models.content import Content
    from llmling_agent.tools.base import Tool


logger = get_logger(__name__)
TResult_co = TypeVar("TResult_co", default=str, covariant=True)
type TNode[TResult_co] = BaseNode | End[FinalResult[TResult_co]]


@dataclass
class ProviderResponse:
    """Raw response data from provider."""

    content: Any
    tool_calls: list[ToolCallInfo] = field(default_factory=list)
    model_name: str = ""
    cost_and_usage: TokenCost | None = None
    provider_extra: dict[str, Any] | None = None


class AgentRunProtocol(Protocol[TResult_co]):
    """Protocol representing an ongoing, iterable agent execution run.

    Matches the interface provided by pydantic_ai.models.AgentRun.
    Allows iterating through execution graph nodes and accessing results/usage.
    """

    def __aiter__(self) -> AsyncIterator[TNode]:
        """Allows `async for node in agent_run:`."""
        ...

    async def __anext__(self) -> TNode:
        """Gets the next node in the iteration."""
        ...

    # --- Manual Iteration Interface ---
    @property
    def next_node(self) -> TNode:
        """The next node that will be run in the agent graph."""
        ...

    async def next(self, node: BaseNode) -> TNode:
        """Manually advances the iteration by running the specified `node`.

        Returns the node *after* the processed one, or an End node.
        """
        ...

    # --- Results and Usage ---
    @property
    def result(self) -> AgentRunResult[TResult_co] | None:
        """The final result (AgentRunResult) of the agent run.

        Available after iteration completes (i.e., an End node is reached).
        Returns None if the run is not finished or failed.
        """
        ...

    def usage(self) -> Usage:
        """Retrieves usage statistics (Usage) for the run so far."""
        ...


@runtime_checkable
class StreamingResponseProtocol[TResult](Protocol):
    """Protocol for streaming responses.

    This matches PydanticAI's StreamedRunResult interface to make transition easier,
    but lives in our core package to remove the dependency.
    """

    model_name: str | None
    is_complete: bool
    formatted_content: TResult = None  # type: ignore

    def stream(self) -> AsyncGenerator[TResult, None]:
        """Stream individual chunks as they arrive."""
        ...

    def stream_text(self, delta: bool = False) -> AsyncGenerator[str, None]:
        """Stream individual chunks as they arrive."""
        ...

    def usage(self) -> Usage:
        """Get token usage statistics if available."""
        ...


@dataclass
class UsageLimits:
    """Limits on model usage."""

    request_limit: int | None = 50
    """The maximum number of requests allowed to the model."""

    input_tokens_limit: int | None = None
    """The maximum number of tokens allowed in requests to the model."""

    output_tokens_limit: int | None = None
    """The maximum number of tokens allowed in responses from the model."""

    total_tokens_limit: int | None = None
    """The maximum number of tokens allowed in requests and responses combined."""


class AgentProvider[TDeps]:
    """Base class for agent providers."""

    tool_used = Signal(ToolCallInfo)
    chunk_streamed = Signal(str, str)
    model_changed = Signal(object)  # Model | None
    NAME: str

    def __init__(
        self,
        *,
        name: str = "agent",
        context: AgentContext[TDeps] | None = None,
        debug: bool = False,
    ):
        super().__init__()
        self._name = name
        self._model: str | ModelProtocol | None = None
        self._context = context
        self._debug = debug

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def set_model(self, model: ModelType):
        """Default no-op implementation for setting model."""

    def iterate_run(
        self,
        *prompts: str | Content,
        message_id: str,
        message_history: list[ChatMessage],
        result_type: type[Any] | None = None,
        model: ModelType = None,
        tools: list[Tool] | None = None,
        usage_limits: UsageLimits | None = None,
        **kwargs: Any,
    ) -> AbstractAsyncContextManager[AgentRunProtocol]:
        """Stream a response. Must be implemented by providers."""
        raise NotImplementedError

    @property
    def context(self) -> AgentContext[TDeps]:
        """Get context."""
        if self._context is None:
            msg = "Context not set"
            raise RuntimeError(msg)
        return self._context

    @context.setter
    def context(self, value: AgentContext[TDeps]):
        """Set context."""
        self._context = value

    @property
    def model_name(self) -> str | None:
        """Get model name."""
        return None

    @property
    def name(self) -> str:
        """Get agent name."""
        return self._name or "agent"

    @name.setter
    def name(self, value: str | None):
        """Set agent name."""
        self._name = value or "agent"

    async def generate_response(
        self,
        *prompts: str | Content,
        message_id: str,
        message_history: list[ChatMessage],
        result_type: type[Any] | None = None,
        model: ModelType = None,
        tools: list[Tool] | None = None,
        usage_limits: UsageLimits | None = None,
        **kwargs: Any,
    ) -> ProviderResponse:
        """Generate a response. Must be implemented by providers."""
        raise NotImplementedError

    def stream_response(
        self,
        *prompts: str | Content,
        message_id: str,
        message_history: list[ChatMessage],
        result_type: type[Any] | None = None,
        model: ModelType = None,
        tools: list[Tool] | None = None,
        usage_limits: UsageLimits | None = None,
        **kwargs: Any,
    ) -> AbstractAsyncContextManager[StreamingResponseProtocol]:
        """Stream a response. Must be implemented by providers."""
        raise NotImplementedError

    async def supports_feature(self, capability: Literal["vision"]) -> bool:
        """Check if provider supports a specific capability."""
        import tokonomics

        match capability:
            case "vision":
                if not self.model_name:
                    return False
                caps = await tokonomics.get_model_capabilities(self.model_name)
                return (caps and caps.supports_vision) or False
            case _:
                return False

    async def get_token_limits(self) -> tokonomics.TokenLimits | None:
        """Get token limits for the current model."""
        import tokonomics

        if not self.model_name:
            return None

        try:
            return await tokonomics.get_model_limits(self.model_name)
        except ValueError:
            logger.debug("Could not get token limits for model: %s", self.model_name)
            return None


class AgentLLMProvider[TDeps](AgentProvider[TDeps]):
    """Provider using LLM backend."""

    def __init__(
        self,
        model_settings: dict[str, Any] | None = None,
        model: str | ModelProtocol | None = None,
        **kwargs: Any,
    ):
        self.model_settings = model_settings or {}
        self._model = model
        super().__init__(**kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(model={self.model_name})"

    def set_model(self, model: ModelType):
        """Set the model for this agent.

        Args:
            model: New model to use (name or instance)

        Emits:
            model_changed signal with the new model
        """
        old_name = self.model_name
        self._model = model
        self.model_changed.emit(model)
        if old_name:
            logger.debug("Changed model from %s to %s", old_name, self.model_name)



"""Initial migration with renamed token fields.

Revision ID: 5ffc5f0266a1
Revises:
Create Date: 2025-09-11 15:08:18.429080

"""

from collections.abc import Sequence

from alembic import op
import sqlalchemy as sa
import sqlmodel.sql.sqltypes

from llmling_agent_storage.sql_provider.models import UTCDateTime


# revision identifiers, used by Alembic.
revision: str = "5ffc5f0266a1"
down_revision: str | Sequence[str] | None = None
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "commandhistory",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("session_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("agent_name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("command", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("context_type", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("context_metadata", sa.JSON(), nullable=True),
        sa.Column("timestamp", UTCDateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_commandhistory_agent_name"),
        "commandhistory",
        ["agent_name"],
        unique=False,
    )
    op.create_index(
        op.f("ix_commandhistory_context_type"),
        "commandhistory",
        ["context_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_commandhistory_session_id"),
        "commandhistory",
        ["session_id"],
        unique=False,
    )
    op.create_table(
        "conversation",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("agent_name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("start_time", UTCDateTime(), nullable=True),
        sa.Column("total_tokens", sa.Integer(), nullable=False),
        sa.Column("total_cost", sa.Float(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_conversation_agent_name"), "conversation", ["agent_name"], unique=False
    )
    op.create_index(
        op.f("ix_conversation_start_time"), "conversation", ["start_time"], unique=False
    )
    op.create_table(
        "message",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("conversation_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("timestamp", UTCDateTime(), nullable=True),
        sa.Column("role", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("content", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("model", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("model_name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("model_provider", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("forwarded_from", sa.JSON(), nullable=True),
        sa.Column("total_tokens", sa.Integer(), nullable=True),
        sa.Column("input_tokens", sa.Integer(), nullable=True),
        sa.Column("output_tokens", sa.Integer(), nullable=True),
        sa.Column("cost", sa.Float(), nullable=True),
        sa.Column("response_time", sa.Float(), nullable=True),
        sa.Column("checkpoint_data", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_message_conversation_id"), "message", ["conversation_id"], unique=False
    )
    op.create_index(op.f("ix_message_cost"), "message", ["cost"], unique=False)
    op.create_index(
        op.f("ix_message_model_name"), "message", ["model_name"], unique=False
    )
    op.create_index(
        op.f("ix_message_model_provider"), "message", ["model_provider"], unique=False
    )
    op.create_index(op.f("ix_message_name"), "message", ["name"], unique=False)
    op.create_index(
        op.f("ix_message_total_tokens"), "message", ["total_tokens"], unique=False
    )
    op.create_table(
        "toolcall",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("conversation_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("message_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("timestamp", UTCDateTime(), nullable=True),
        sa.Column("tool_call_id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("tool_name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("args", sa.JSON(), nullable=True),
        sa.Column("result", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_toolcall_conversation_id"), "toolcall", ["conversation_id"], unique=False
    )
    op.create_index(
        op.f("ix_toolcall_message_id"), "toolcall", ["message_id"], unique=False
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_toolcall_message_id"), table_name="toolcall")
    op.drop_index(op.f("ix_toolcall_conversation_id"), table_name="toolcall")
    op.drop_table("toolcall")
    op.drop_index(op.f("ix_message_total_tokens"), table_name="message")
    op.drop_index(op.f("ix_message_name"), table_name="message")
    op.drop_index(op.f("ix_message_model_provider"), table_name="message")
    op.drop_index(op.f("ix_message_model_name"), table_name="message")
    op.drop_index(op.f("ix_message_cost"), table_name="message")
    op.drop_index(op.f("ix_message_conversation_id"), table_name="message")
    op.drop_table("message")
    op.drop_index(op.f("ix_conversation_start_time"), table_name="conversation")
    op.drop_index(op.f("ix_conversation_agent_name"), table_name="conversation")
    op.drop_table("conversation")
    op.drop_index(op.f("ix_commandhistory_session_id"), table_name="commandhistory")
    op.drop_index(op.f("ix_commandhistory_context_type"), table_name="commandhistory")
    op.drop_index(op.f("ix_commandhistory_agent_name"), table_name="commandhistory")
    op.drop_table("commandhistory")
    # ### end Alembic commands ###



"""McpRun based toolset implementation."""

from __future__ import annotations

import os

from llmling.core.log import get_logger

from llmling_agent.resource_providers.base import ResourceProvider
from llmling_agent.tools.base import Tool


logger = get_logger(__name__)


class McpRunTools(ResourceProvider):
    """Provider for composio tools."""

    def __init__(self, entity_id: str, session_id: str | None = None):
        from mcp_run import Client, ClientConfig

        super().__init__(name=entity_id)
        id_ = session_id or os.environ.get("MCP_RUN_SESSION_ID")
        config = ClientConfig()
        self.client = Client(session_id=id_, config=config)
        self._tools: list[Tool] | None = None

    async def get_tools(self) -> list[Tool]:
        """Get tools from entry points."""
        # Return cached tools if available
        if self._tools is not None:
            return self._tools

        self._tools = []
        for name, tool in self.client.tools.items():

            def run(tool_name=name, **input_dict):
                return self.client.call_tool(tool_name, params=input_dict)

            run.__name__ = name

            tool = Tool.from_callable(run, schema_override=tool.input_schema)  # pyright: ignore
            self._tools.append(tool)

        return self._tools


if __name__ == "__main__":
    import asyncio

    async def main():
        tools = McpRunTools("default")
        fns = await tools.get_tools()
        print(fns)

    asyncio.run(main())



#!/usr/bin/env python3
"""Database migration management script for llmling-agent.

This script provides easy commands for managing database migrations using Alembic.
"""

from __future__ import annotations

import argparse
import os
from pathlib import Path
import subprocess
import sys


def run_command(cmd: list[str], cwd: Path | None = None) -> int:
    """Run a command and return the exit code."""
    print(f"Running: {' '.join(cmd)}")
    result = subprocess.run(cmd, cwd=cwd, check=False)
    if result.returncode != 0 and cmd[0] == "alembic":
        print(f"Error: Command '{cmd[0]}' not found. Make sure alembic is installed.")
    return result.returncode


def get_project_root() -> Path:
    """Get the project root directory."""
    # Start from this file and go up to find the project root
    # This file is in src/llmling_agent_storage/sql_provider/cli.py
    return Path(__file__).parent.parent.parent.parent


def main():  # noqa: PLR0911
    """Main entry point."""
    parser = argparse.ArgumentParser(
        description="Database migration management for llmling-agent"
    )

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Status command
    subparsers.add_parser("status", help="Show current migration status")

    # Upgrade command
    upgrade_parser = subparsers.add_parser(
        "upgrade", help="Upgrade database to latest migration"
    )
    upgrade_parser.add_argument(
        "revision", nargs="?", default="head", help="Target revision (default: head)"
    )

    # Downgrade command
    downgrade_parser = subparsers.add_parser("downgrade", help="Downgrade database")
    downgrade_parser.add_argument(
        "revision",
        help="Target revision (e.g., -1 for previous, or specific revision ID)",
    )

    # Create migration command
    create_parser = subparsers.add_parser("create", help="Create a new migration")
    create_parser.add_argument("message", help="Migration message")
    create_parser.add_argument(
        "--autogenerate",
        action="store_true",
        help="Auto-generate migration based on model changes",
    )

    # History command
    subparsers.add_parser("history", help="Show migration history")

    # Current command
    subparsers.add_parser("current", help="Show current revision")

    # Reset command
    reset_parser = subparsers.add_parser(
        "reset", help="Reset database (WARNING: destructive)"
    )
    reset_parser.add_argument(
        "--force", action="store_true", help="Force reset without confirmation"
    )

    args = parser.parse_args()

    if not args.command:
        parser.print_help()
        return 1

    project_root = get_project_root()

    # Set up environment
    env = os.environ.copy()
    env["PYTHONPATH"] = str(project_root / "src")

    # Handle commands
    if args.command == "status":
        return run_command(["alembic", "current", "-v"], cwd=project_root)

    if args.command == "upgrade":
        return run_command(["alembic", "upgrade", args.revision], cwd=project_root)

    if args.command == "downgrade":
        return run_command(["alembic", "downgrade", args.revision], cwd=project_root)

    if args.command == "history":
        return run_command(["alembic", "history", "-v"], cwd=project_root)

    if args.command == "current":
        return run_command(["alembic", "current"], cwd=project_root)

    if args.command == "create":
        cmd = ["alembic", "revision", "-m", args.message]
        if args.autogenerate:
            cmd.append("--autogenerate")
        return run_command(cmd, cwd=project_root)

    if args.command == "reset":
        if not args.force:
            response = input(
                "This will delete all data in the database. Continue? (y/N): "
            )
            if response.lower() != "y":
                print("Aborted.")
                return 0

        # Drop all tables and recreate
        print("Resetting database...")

        # First, drop to base (empty database)
        result = run_command(["alembic", "downgrade", "base"], cwd=project_root)
        if result != 0:
            return result

        # Then upgrade to head
        return run_command(["alembic", "upgrade", "head"], cwd=project_root)

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Models for OpenAI responses endpoint."""

from __future__ import annotations

from collections.abc import Sequence  # noqa: TC003
from datetime import datetime
from typing import Any, Literal, TypedDict
from uuid import uuid4

from pydantic import Field
from schemez import Schema


class InputText(Schema):
    """Text input part."""

    type: Literal["input_text"] = "input_text"
    text: str


class InputImage(Schema):
    """Image input part."""

    type: Literal["input_image"] = "input_image"
    image_url: str


class ResponseOutputText(Schema):
    """Text output part."""

    type: Literal["output_text"] = "output_text"
    text: str
    annotations: list[dict[str, Any]] = Field(default_factory=list)


class ResponseToolCall(Schema):
    """Tool call in response."""

    type: str  # web_search_call etc
    id: str
    status: Literal["completed", "error"] = "completed"


class ResponseMessage(Schema):
    """ResponseMessage in response."""

    type: Literal["message"] = "message"
    id: str
    status: Literal["completed", "error"] = "completed"
    role: Literal["user", "assistant", "system"]
    content: list[ResponseOutputText]


class ResponseUsage(TypedDict):
    """Token usage information."""

    input_tokens: int
    input_tokens_details: dict[str, int]
    output_tokens: int
    output_tokens_details: dict[str, int]
    total_tokens: int


class ResponseRequest(Schema):
    """Request for /v1/responses endpoint."""

    model: str
    input: str | list[dict[str, Any]]
    instructions: str | None = None
    stream: bool = False
    temperature: float = 1.0
    tools: list[dict[str, Any]] = Field(default_factory=list)
    tool_choice: str = "auto"
    max_output_tokens: int | None = None
    metadata: dict[str, str] = Field(default_factory=dict)


class Response(Schema):
    """Response from /v1/responses endpoint."""

    id: str = Field(default_factory=lambda: f"resp_{uuid4().hex}")
    object: Literal["response"] = "response"
    created_at: int = Field(default_factory=lambda: int(datetime.now().timestamp()))
    status: Literal["completed", "error"] = "completed"
    error: str | None = None
    model: str
    output: Sequence[ResponseMessage | ResponseToolCall]

    # Include all the request parameters
    instructions: str | None = None
    max_output_tokens: int | None = None
    temperature: float = 1.0
    tools: list[dict[str, Any]] = Field(default_factory=list)
    tool_choice: str = "auto"
    usage: ResponseUsage | None = None
    metadata: dict[str, str] = Field(default_factory=dict)



"""Agent provider implementations."""

from __future__ import annotations

from contextlib import asynccontextmanager
from dataclasses import asdict
from functools import wraps
from typing import TYPE_CHECKING, Any, cast, get_args

from llmling_models import AllModels, infer_model
from pydantic_ai import Agent as PydanticAgent
import pydantic_ai._function_schema
from pydantic_ai.messages import ModelResponse
from pydantic_ai.models import KnownModelName, Model
from pydantic_ai.result import StreamedRunResult
from pydantic_ai.tools import GenerateToolJsonSchema, RunContext
from pydantic_ai.usage import UsageLimits as PydanticAiUsageLimits

from llmling_agent.agent.context import AgentContext
from llmling_agent.common_types import EndStrategy, ModelProtocol
from llmling_agent.log import get_logger
from llmling_agent.messaging.messages import ChatMessage, TokenCost
from llmling_agent.observability import track_action
from llmling_agent.tasks.exceptions import (
    ChainAbortedError,
    RunAbortedError,
    ToolSkippedError,
)
from llmling_agent.utils.inspection import execute, has_argument_type
from llmling_agent_providers.base import (
    AgentLLMProvider,
    AgentRunProtocol,
    ProviderResponse,
    UsageLimits,
)
from llmling_agent_providers.pydanticai.utils import (
    convert_prompts_to_user_content,
    format_part,
    get_tool_calls,
    to_model_message,
)


if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Awaitable, Callable

    from pydantic_ai.agent import AgentRunResult

    from llmling_agent.common_types import ModelType
    from llmling_agent.models.content import Content
    from llmling_agent.tools.base import Tool


logger = get_logger(__name__)


# 🤫 Secret agent stuff below
# We're doing some "creative" context handling here to make pydantic-ai accept our
# AgentContext.
# If you're reading this, please don't judge too harshly.
# Sometimes you gotta do what you gotta do to keep the architecture clean(ish).


def _is_call_ctx(annotation: Any) -> bool:
    # Yes, we're monkey-patching pydantic-ai's internal function.
    # No, we're not proud of it.
    # But it works™
    from typing import get_origin

    from pydantic._internal import _typing_extra
    from pydantic_ai.tools import RunContext

    if annotation is RunContext or (
        _typing_extra.is_generic_alias(annotation)
        and get_origin(annotation) is RunContext
    ):
        return True
    return annotation is AgentContext or (
        _typing_extra.is_generic_alias(annotation)
        and get_origin(annotation) is AgentContext
    )


pydantic_ai._function_schema._is_call_ctx = _is_call_ctx


class PydanticAIProvider(AgentLLMProvider):
    """Provider using pydantic-ai as backend."""

    NAME = "pydantic_ai"

    def __init__(
        self,
        *,
        model: str | ModelProtocol | None = None,
        context: AgentContext | None = None,
        name: str = "agent",
        retries: int = 1,
        output_retries: int | None = None,
        end_strategy: EndStrategy = "early",
        defer_model_check: bool = False,
        debug: bool = False,
        model_settings: dict[str, Any] | None = None,
    ):
        """Initialize pydantic-ai backend.

        Args:
            model: Model to use for responses
            context: Agent context
            name: Agent name
            retries: Number of retries for failed operations
            output_retries: Max retries for result validation
            end_strategy: How to handle tool calls with final result
            defer_model_check: Whether to defer model validation
            debug: Whether to enable debug mode
            model_settings: Additional model-specific settings
        """
        super().__init__(
            model=model,
            model_settings=model_settings,
            debug=debug,
            context=context,
        )
        self._kwargs: dict[str, Any] = dict(
            model=model,
            name=name,
            tools=[],
            retries=retries,
            end_strategy=end_strategy,
            output_retries=output_retries,
            defer_model_check=defer_model_check,
        )

    async def get_model_names(self) -> list[str]:
        """Get list of all known model names."""
        return list(get_args(KnownModelName)) + list(get_args(AllModels))

    async def get_agent(
        self,
        system_prompt: str,
        tools: list[Tool],
    ) -> PydanticAgent[Any, Any]:
        kwargs = self._kwargs.copy()
        model = kwargs.pop("model", None)
        model = infer_model(model) if isinstance(model, str) else model
        agent = PydanticAgent(model=model, instructions=system_prompt, **kwargs)  # type: ignore
        for tool in tools:
            wrapped = (
                self.wrap_tool(tool, self._context)
                if self._context
                else tool.callable.callable
            )
            if has_argument_type(wrapped, RunContext):
                agent.tool(wrapped)
            elif has_argument_type(wrapped, AgentContext):
                agent._function_toolset.add_function(
                    wrapped,
                    True,
                    None,
                    1,
                    None,
                    "auto",
                    False,
                    GenerateToolJsonSchema,
                    None,
                )
            else:
                agent.tool_plain(wrapped)
        return agent

    @property
    def model(self) -> str | ModelType:
        return self._model

    def wrap_tool(
        self,
        tool: Tool,
        agent_ctx: AgentContext,
    ) -> Callable[..., Awaitable[Any]]:
        """Wrap tool with confirmation handling.

        We wrap the tool to intercept pydantic-ai's tool calls and add our confirmation
        logic before the actual execution happens. The actual tool execution (including
        moving sync functions to threads) is handled by pydantic-ai.

        Current situation is: We only get all infos for tool calls for functions with
        RunContext. In order to migitate this, we "fallback" to the AgentContext, which
        at least provides some information.
        """
        original_tool = tool.callable.callable
        if has_argument_type(original_tool, RunContext):

            async def wrapped(ctx: RunContext[AgentContext], *args, **kwargs):  # pyright: ignore
                result = await agent_ctx.handle_confirmation(tool, kwargs)
                if agent_ctx.report_progress:
                    await agent_ctx.report_progress(ctx.run_step, None)
                match result:
                    case "allow":
                        return await execute(original_tool, ctx, *args, **kwargs)
                    case "skip":
                        msg = f"Tool {tool.name} execution skipped"
                        raise ToolSkippedError(msg)
                    case "abort_run":
                        msg = "Run aborted by user"
                        raise RunAbortedError(msg)
                    case "abort_chain":
                        msg = "Agent chain aborted by user"
                        raise ChainAbortedError(msg)

        elif has_argument_type(original_tool, AgentContext):

            async def wrapped(ctx: AgentContext, *args, **kwargs):  # pyright: ignore
                result = await agent_ctx.handle_confirmation(tool, kwargs)
                match result:
                    case "allow":
                        return await execute(original_tool, agent_ctx, *args, **kwargs)
                    case "skip":
                        msg = f"Tool {tool.name} execution skipped"
                        raise ToolSkippedError(msg)
                    case "abort_run":
                        msg = "Run aborted by user"
                        raise RunAbortedError(msg)
                    case "abort_chain":
                        msg = "Agent chain aborted by user"
                        raise ChainAbortedError(msg)

        else:

            async def wrapped(*args, **kwargs):  # pyright: ignore
                result = await agent_ctx.handle_confirmation(tool, kwargs)
                match result:
                    case "allow":
                        return await execute(original_tool, *args, **kwargs)
                    case "skip":
                        msg = f"Tool {tool.name} execution skipped"
                        raise ToolSkippedError(msg)
                    case "abort_run":
                        msg = "Run aborted by user"
                        raise RunAbortedError(msg)
                    case "abort_chain":
                        msg = "Agent chain aborted by user"
                        raise ChainAbortedError(msg)

        wraps(original_tool)(wrapped)  # pyright: ignore
        wrapped.__doc__ = tool.description
        wrapped.__name__ = tool.name
        return wrapped

    @track_action("Pydantic-AI call. model: {model} result type {result_type}.")
    async def generate_response(
        self,
        *prompts: str | Content,
        message_id: str,
        message_history: list[ChatMessage],
        result_type: type[Any] | None = None,
        model: ModelType = None,
        tools: list[Tool] | None = None,
        system_prompt: str | None = None,
        usage_limits: UsageLimits | None = None,
        **kwargs: Any,
    ) -> ProviderResponse:
        """Generate response using pydantic-ai."""
        agent = await self.get_agent(system_prompt or "", tools=tools or [])
        use_model = model or self.model
        if isinstance(use_model, str):
            use_model = infer_model(use_model)
            self.model_changed.emit(use_model)
        try:
            # Convert prompts to pydantic-ai format
            converted_prompts = await convert_prompts_to_user_content(prompts)

            # Run with complete history
            to_use = model or self.model
            to_use = infer_model(to_use) if isinstance(to_use, str) else to_use
            limits = asdict(usage_limits) if usage_limits else {}
            result: AgentRunResult = await agent.run(
                converted_prompts,  # Pass converted prompts
                deps=self._context,  # type: ignore
                message_history=[to_model_message(m) for m in message_history],
                model=to_use,  # type: ignore
                output_type=result_type or str,
                model_settings=self.model_settings,  # type: ignore
                usage_limits=PydanticAiUsageLimits(**limits),
            )

            # Extract tool calls and set message_id
            new_msgs = result.new_messages()
            tool_dict = {i.name: i for i in tools or []}
            tool_calls = get_tool_calls(new_msgs, tool_dict, agent_name=self.name)
            for call in tool_calls:
                call.message_id = message_id
                call.context_data = self._context.data if self._context else None
                self.tool_used.emit(call)

            resolved_model = (
                use_model.model_name if isinstance(use_model, Model) else str(use_model)
            )
            usage = result.usage()
            # Create input content representation for cost calculations
            cost_input = "\n".join(str(p) for p in prompts)
            cost_info = (
                await TokenCost.from_usage(
                    usage,
                    resolved_model,
                    cost_input,
                    str(result.output),
                )
                if resolved_model and usage
                else None
            )
            return ProviderResponse(
                content=result.output,
                tool_calls=tool_calls,
                cost_and_usage=cost_info,
                model_name=resolved_model,
            )
        finally:
            # Restore original model in signal if we had an override
            if model:
                original = self.model
                if isinstance(original, str):
                    original = infer_model(original)
                self.model_changed.emit(original)

    @property
    def name(self) -> str:
        """Get agent name."""
        return self._kwargs.get("name", "agent")  # type: ignore

    @name.setter
    def name(self, value: str | None):
        """Set agent name."""
        self._kwargs["name"] = value

    def set_model(self, model: ModelType):
        """Set the model for this agent.

        Args:
            model: New model to use (name or instance)

        Emits:
            model_changed signal with the new model
        """
        old_name = self.model_name
        if isinstance(model, str):
            model = infer_model(model)
        self._model = model
        self._kwargs["model"] = model
        self.model_changed.emit(model)
        logger.debug("Changed model from %s to %s", old_name, self.model_name)

    @property
    def model_name(self) -> str | None:
        """Get the model name in a consistent format."""
        match model := self._kwargs["model"]:
            case str() | None:
                return model
            case ModelProtocol():
                return model.model_name
            case _:
                msg = f"Invalid model type: {model}"
                raise ValueError(msg)

    @asynccontextmanager
    async def iterate_run(
        self,
        *prompts: str | Content,  # Provider signature
        message_id: str,
        message_history: list[ChatMessage[Any]],  # Use Any generic
        tools: list[Tool] | None = None,
        result_type: type[Any] | None = None,
        usage_limits: UsageLimits | None = None,
        model: ModelType = None,
        system_prompt: str | None = None,
        **kwargs: Any,  # Analogous kwargs
    ) -> AsyncIterator[AgentRunProtocol[Any]]:  # Use Any generic for now
        """Starts an iterable agent run using pydantic-ai's Agent.iter."""
        agent = await self.get_agent(system_prompt or "", tools=tools or [])
        use_model = model or self.model
        original_model_signal = None
        if model:
            original_model_signal = self.model
            if isinstance(use_model, str):
                use_model = infer_model(use_model)
            # Don't emit model_changed here, agent.iter takes model directly
            # self.model_changed.emit(use_model) # NO EMIT NEEDED HERE
        elif isinstance(use_model, str):
            use_model = infer_model(use_model)

        limits = asdict(usage_limits) if usage_limits else {}
        pydantic_limits = PydanticAiUsageLimits(**limits)
        resolved_model_name = (
            use_model.model_name if isinstance(use_model, Model) else str(use_model)
        )

        try:
            converted_prompts = await convert_prompts_to_user_content(prompts)
            model_messages = [to_model_message(m) for m in message_history]
            msg = "Starting PydanticAI agent iteration run_id=%s with model=%s"
            logger.debug(msg, message_id, resolved_model_name)
            async with agent.iter(
                converted_prompts,
                deps=self._context,
                message_history=model_messages,
                model=model or self.model,  # type: ignore
                output_type=result_type or str,
                model_settings=self.model_settings,  # type: ignore
                usage_limits=pydantic_limits,
                **kwargs,
            ) as agent_run:
                # 7. Post-processing (Minimal for iter, just yield)
                # Add model name to the run object if possible/needed?
                # setattr(agent_run, 'model_name', resolved_model_name)

                yield cast(AgentRunProtocol[Any], agent_run)  # Yield the run object
        except Exception as e:
            # Catch potential errors from argument mismatches in agent.iter
            logger.exception("Error calling agent.iter")
            msg = f"Iteration setup error: {e}"
            raise RuntimeError(msg) from e
        finally:
            if original_model_signal and isinstance(original_model_signal, str):
                original_model_signal = infer_model(original_model_signal)
            # self.model_changed.emit(original_model_signal)

    @asynccontextmanager
    async def stream_response(  # type: ignore[override]
        self,
        *prompts: str | Content,
        message_history: list[ChatMessage],
        message_id: str,
        result_type: type[Any] | None = None,
        model: ModelType = None,
        tools: list[Tool] | None = None,
        system_prompt: str | None = None,
        usage_limits: UsageLimits | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[StreamedRunResult]:  # type: ignore[type-var]
        """Stream response using pydantic-ai."""
        agent = await self.get_agent(system_prompt or "", tools=tools or [])
        use_model = model or self.model
        limits = asdict(usage_limits) if usage_limits else {}
        if isinstance(use_model, str):
            use_model = infer_model(use_model)

        if model:
            self.model_changed.emit(use_model)

        # Convert prompts to pydantic-ai format
        converted_prompts = await convert_prompts_to_user_content(prompts)

        # Convert all messages to pydantic-ai format
        model_messages = [to_model_message(m) for m in message_history]

        async with agent.run_stream(
            converted_prompts,
            deps=self._context,
            message_history=model_messages,
            model=model or self.model,  # type: ignore
            output_type=result_type or str,
            model_settings=self.model_settings,  # type: ignore
            usage_limits=PydanticAiUsageLimits(**limits),
        ) as stream_result:
            stream_result = cast(StreamedRunResult[AgentContext[Any], Any], stream_result)
            original_stream = stream_result.stream
            original_text_stream = stream_result.stream_text
            resolved_model = (
                use_model.model_name if isinstance(use_model, Model) else str(use_model)
            )
            stream_result.model_name = resolved_model  # type: ignore

            def get_wrapped_stream(fn):
                async def wrapped_stream(*args, **kwargs):
                    last_content = None
                    async for chunk in fn(*args, **kwargs):
                        # Only emit if content has changed
                        if chunk != last_content:
                            self.chunk_streamed.emit(str(chunk), message_id)
                            last_content = chunk
                            yield chunk

                    if stream_result.is_complete:
                        self.chunk_streamed.emit("", message_id)
                        messages = stream_result.new_messages()
                        tool_dict = {i.name: i for i in tools or []}
                        # Extract and update tool calls
                        tool_calls = get_tool_calls(
                            messages, tool_dict, agent_name=self.name
                        )
                        for call in tool_calls:
                            call.message_id = message_id
                            call.context_data = (
                                self._context.data if self._context else None
                            )
                            self.tool_used.emit(call)
                        # Format final content
                        responses = [m for m in messages if isinstance(m, ModelResponse)]
                        parts = [p for msg in responses for p in msg.parts]
                        content = "\n".join(format_part(p) for p in parts)
                        # Update stream result with formatted content
                        stream_result.formatted_content = content  # type: ignore

                return wrapped_stream

            if model:
                original = self.model
                if isinstance(original, str):
                    original = infer_model(original)
                self.model_changed.emit(original)

            stream_result.stream = get_wrapped_stream(original_stream)  # type: ignore
            stream_result.stream_text = get_wrapped_stream(original_text_stream)  # type: ignore
            yield stream_result  # type: ignore



"""Callback input provider implementation."""

from __future__ import annotations

from collections.abc import AsyncIterator, Awaitable, Callable, Iterator
from typing import TYPE_CHECKING, Any

from llmling_agent.utils.inspection import execute
from llmling_agent_input.base import InputProvider


if TYPE_CHECKING:
    from llmling_agent.agent.context import AgentContext, ConfirmationResult
    from llmling_agent.messaging.messages import ChatMessage
    from llmling_agent.tools.base import Tool


class CallbackInputProvider(InputProvider):
    """Input provider that delegates to provided callbacks."""

    def __init__(
        self,
        get_input: Callable[..., str | Awaitable[str]] | None = None,
        get_streaming_input: Callable[..., AsyncIterator[str] | Iterator[str]]
        | None = None,
        get_tool_confirmation: Callable[
            ..., ConfirmationResult | Awaitable[ConfirmationResult]
        ]
        | None = None,
        get_code_input: Callable[..., str | Awaitable[str]] | None = None,
    ):
        self._get_input = get_input
        self._get_streaming = get_streaming_input
        self._get_confirmation = get_tool_confirmation
        self._get_code = get_code_input

    async def get_input(
        self,
        context: AgentContext,
        prompt: str,
        result_type: type | None = None,
        message_history: list[ChatMessage] | None = None,
    ) -> Any:
        if not self._get_input:
            return input(prompt)  # fallback
        return await execute(
            self._get_input,
            context=context,
            prompt=prompt,
            result_type=result_type,
            message_history=message_history,
        )

    async def _get_streaming_input(  # type: ignore
        self,
        context: AgentContext,
        prompt: str,
        result_type: type | None = None,
        message_history: list[ChatMessage] | None = None,
    ) -> AsyncIterator[str]:
        if not self._get_streaming:
            # Use parent class fallback
            return await super().get_streaming_input(  # type: ignore
                context, prompt, result_type, message_history
            )

        iterator = self._get_streaming(
            context=context,
            prompt=prompt,
            result_type=result_type,
            message_history=message_history,
        )

        if isinstance(iterator, AsyncIterator):
            return iterator

        async def wrap_sync():  # wrap sync iterator
            for item in iterator:
                yield item

        return wrap_sync()

    async def get_tool_confirmation(
        self,
        context: AgentContext,
        tool: Tool,
        args: dict[str, Any],
        message_history: list[ChatMessage] | None = None,
    ) -> ConfirmationResult:
        if not self._get_confirmation:
            return "allow"  # fallback: always allow
        return await execute(
            self._get_confirmation,
            context=context,
            tool=tool,
            args=args,
            message_history=message_history,
        )

    async def get_code_input(
        self,
        context: AgentContext,
        template: str | None = None,
        language: str = "python",
        description: str | None = None,
    ) -> str:
        if not self._get_code:
            return input("Enter code: ")  # basic fallback
        return await execute(
            self._get_code,
            context=context,
            template=template,
            language=language,
            description=description,
        )



"""Helpers for OpenAI-compatible API server."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

import anyenv

from llmling_agent.log import get_logger


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from llmling_agent import AnyAgent
    from llmling_agent_server.models import ChatCompletionRequest

logger = get_logger(__name__)


async def stream_response(
    agent: AnyAgent[Any, Any],
    content: str,
    request: ChatCompletionRequest,
) -> AsyncGenerator[str, None]:
    """Generate streaming response chunks."""
    response_id = f"chatcmpl-{int(time.time() * 1000)}"
    created = int(time.time())

    try:
        # First chunk with role
        choice = {"index": 0, "delta": {"role": "assistant"}, "finish_reason": None}
        first_chunk = {
            "id": response_id,
            "object": "chat.completion.chunk",
            "created": created,
            "model": request.model,
            "choices": [choice],
        }
        yield f"data: {anyenv.dump_json(first_chunk)}\n\n"
        async with agent.run_stream(content) as stream:
            async for chunk in stream.stream_text(delta=True):
                # Skip empty chunks
                if not chunk:
                    continue
                choice = {
                    "index": 0,
                    "delta": {"content": chunk},
                    "finish_reason": None,
                }
                chunk_data = {
                    "id": response_id,
                    "object": "chat.completion.chunk",
                    "created": created,
                    "model": request.model,
                    "choices": [choice],
                }
                yield f"data: {anyenv.dump_json(chunk_data)}\n\n"
        final_chunk = {
            "id": response_id,
            "object": "chat.completion.chunk",
            "created": created,
            "model": request.model,
            "choices": [{"index": 0, "delta": {}, "finish_reason": "stop"}],
        }
        yield f"data: {anyenv.dump_json(final_chunk)}\n\n"
        yield "data: [DONE]\n\n"

    except Exception as e:
        logger.exception("Error during streaming response")
        choice = {
            "index": 0,
            "delta": {"content": f"Error: {e!s}"},
            "finish_reason": "error",
        }
        error_chunk = {
            "id": response_id,
            "object": "chat.completion.chunk",
            "created": created,
            "model": request.model,
            "choices": [choice],
        }
        yield f"data: {anyenv.dump_json(error_chunk)}\n\n"
        yield "data: [DONE]\n\n"



"""Converters for Content types to LiteLLM formats."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any
import warnings

from llmling_agent.models import content as own_content


if TYPE_CHECKING:
    from llmling_agent.models.content import Content


def content_to_litellm_format(content: Content) -> dict[str, Any] | None:  # noqa: PLR0911
    """Convert our Content types to LiteLLM-compatible format.

    Args:
        content: One of our Content instances to convert

    Returns:
        Converted LiteLLM-compatible content dict or None if conversion not possible

    Note:
        Some content types may not have an equivalent in LiteLLM.
        In these cases, None is returned and a warning is emitted.
    """
    match content:
        case own_content.ImageURLContent():
            return {
                "type": "image_url",
                "image_url": {"url": content.url, "detail": content.detail or "auto"},
            }

        case own_content.ImageBase64Content():
            data_url = f"data:{content.mime_type};base64,{content.data}"
            return {
                "type": "image_url",
                "image_url": {
                    "url": data_url,
                    "detail": content.detail or "auto",
                    "format": content.mime_type,
                },
            }

        case own_content.PDFURLContent():
            # LiteLLM treats PDFs as image_url type
            return {
                "type": "image_url",
                "image_url": {
                    "url": content.url,
                    "detail": content.detail or "auto",
                    "format": "application/pdf",
                },
            }

        case own_content.PDFBase64Content():
            # LiteLLM treats PDFs as image_url type
            data_url = f"data:application/pdf;base64,{content.data}"
            return {
                "type": "image_url",
                "image_url": {
                    "url": data_url,
                    "detail": content.detail or "auto",
                    "format": "application/pdf",
                },
            }

        case own_content.AudioURLContent():
            return {
                "type": "input_audio",
                "input_audio": {"url": content.url, "format": content.format or "mp3"},
            }

        case own_content.AudioBase64Content():
            return {
                "type": "input_audio",
                "input_audio": {"data": content.data, "format": content.format or "mp3"},
            }

        case own_content.VideoURLContent():
            msg = "Video content not consistently supported by LiteLLM models"
            warnings.warn(msg, UserWarning, stacklevel=2)
            return {
                "type": "video_url",
                "video_url": {"url": content.url, "format": content.format},
            }

        case _:
            msg = f"Unknown content type: {type(content)}"
            warnings.warn(msg, UserWarning, stacklevel=2)
            return None



import asyncio
from logging.config import fileConfig
import os

from alembic import context
from sqlalchemy import pool
from sqlalchemy.engine import Connection
from sqlalchemy.ext.asyncio import async_engine_from_config
from sqlmodel import SQLModel

from llmling_agent_storage.sql_provider.models import *  # noqa: F403


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)


# Set target metadata from SQLModel
target_metadata = SQLModel.metadata

# Set database URL from environment variable if available
if not config.get_main_option("sqlalchemy.url"):
    database_url = os.getenv("DATABASE_URL", "sqlite:///./llmling_agent.db")
    config.set_main_option("sqlalchemy.url", database_url)


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        compare_type=True,
        compare_server_default=True,
    )

    with context.begin_transaction():
        context.run_migrations()


def do_run_migrations(connection: Connection) -> None:
    context.configure(
        connection=connection,
        target_metadata=target_metadata,
        compare_type=True,
        compare_server_default=True,
    )

    with context.begin_transaction():
        context.run_migrations()


async def run_async_migrations() -> None:
    """Run migrations for async database."""
    connectable = async_engine_from_config(
        config.get_section(config.config_ini_section, {}),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    async with connectable.connect() as connection:
        await connection.run_sync(do_run_migrations)

    await connectable.dispose()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    # Check if we're using an async engine
    url = config.get_main_option("sqlalchemy.url")
    if url and ("sqlite+aiosqlite" in url or "postgresql+asyncpg" in url):
        asyncio.run(run_async_migrations())
    else:
        # Synchronous migration
        from sqlalchemy import engine_from_config

        connectable = engine_from_config(
            config.get_section(config.config_ini_section, {}),
            prefix="sqlalchemy.",
            poolclass=pool.NullPool,
        )

        with connectable.connect() as connection:
            do_run_migrations(connection)


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""Agent injection utilities."""

from __future__ import annotations

import inspect
import typing
from typing import TYPE_CHECKING, Any

from llmling_agent.log import get_logger


if TYPE_CHECKING:
    from collections.abc import Callable

    from llmling_agent import AgentPool, MessageNode


logger = get_logger(__name__)


def is_node_type(typ: Any) -> bool:
    """Check if a type is or inherits from MessageNode."""
    from llmling_agent import MessageNode

    # Direct type check
    if typ is MessageNode:
        return True

    # For actual types (not GenericAlias etc)
    if isinstance(typ, type):
        return issubclass(typ, MessageNode)

    # For generic types (Agent[T], StructuredAgent[T], etc)
    origin = getattr(typ, "__origin__", None)
    if origin is not None and isinstance(origin, type):
        return issubclass(origin, MessageNode)

    return False


class NodeInjectionError(Exception):
    """Raised when agent injection fails."""


def inject_nodes[T, **P](
    func: Callable[P, T],
    pool: AgentPool,
    provided_kwargs: dict[str, Any],
) -> dict[str, MessageNode[Any, Any]]:
    """Get nodes to inject based on function signature."""
    hints = typing.get_type_hints(func)
    params = inspect.signature(func).parameters
    msg = "Injecting nodes for %s.%s"
    logger.debug(msg, func.__module__, func.__qualname__)
    logger.debug("Type hints: %s", hints)
    logger.debug("Available nodes in pool: %s", sorted(pool.nodes))

    nodes: dict[str, MessageNode[Any, Any]] = {}
    for name, param in params.items():
        if param.kind not in (
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.KEYWORD_ONLY,
        ):
            logger.debug("Skipping %s: wrong parameter kind %s", name, param.kind)
            continue

        hint = hints.get(name)
        if hint is None:
            logger.debug("Skipping %s: no type hint", name)
            continue

        # Handle Optional/Union types
        origin = getattr(hint, "__origin__", None)
        args = getattr(hint, "__args__", ())

        # Check for MessageNode or any of its subclasses
        is_node = (
            is_node_type(hint)  # Direct node type
            or (  # Optional[Node[T]] or Union containing Node
                origin is not None and any(is_node_type(arg) for arg in args)
            )
        )

        if not is_node:
            msg = "Skipping %s: not a node type (hint=%s, origin=%s, args=%s)"
            logger.debug(msg, name, hint, origin, args)
            continue

        logger.debug("Found node parameter: %s", name)

        # Check for duplicate parameters
        if name in provided_kwargs and provided_kwargs[name] is not None:
            msg = (
                f"Cannot inject node '{name}': Parameter already provided.\n"
                f"Remove the explicit argument or rename the parameter."
            )
            logger.error(msg)
            raise NodeInjectionError(msg)

        # Get node from pool
        if name not in pool.nodes:
            available = ", ".join(sorted(pool.nodes))
            msg = (
                f"No node named '{name}' found in pool.\n"
                f"Available nodes: {available}\n"
                f"Check your YAML configuration or node name."
            )
            logger.error(msg)
            raise NodeInjectionError(msg)

        nodes[name] = pool.nodes[name]
        logger.debug("Injecting node %s for parameter %s", nodes[name], name)

    logger.debug("Injection complete. Injected nodes: %s", sorted(nodes))
    return nodes



"""Slashed command to list nodes."""

from __future__ import annotations

from slashed import CommandContext, SlashedCommand

from llmling_agent.log import get_logger
from llmling_agent.messaging.context import NodeContext  # noqa: TC001


logger = get_logger(__name__)


class ListNodesCommand(SlashedCommand):
    """List all nodes in the pool with their status."""

    name = "list-nodes"
    category = "pool"

    async def execute_command(
        self,
        ctx: CommandContext[NodeContext],
        show_connections: bool = False,
    ):
        """List all nodes and their current status.

        Args:
            ctx: Command context with node
            show_connections: Whether to show node connections
        """
        node = ctx.get_data()
        header = "\nAvailable Nodes:"
        lines = [header, "=" * len(header)]
        assert node.pool
        for name, node_ in node.pool.nodes.items():
            # Build status info
            status = "🔄 busy" if node_.is_busy() else "⏳ idle"

            # Add connections if requested
            connections = []
            if show_connections and node_.connections.get_targets():
                connections = [a.name for a in node_.connections.get_targets()]
                conn_str = f" → {', '.join(connections)}"
            else:
                conn_str = ""

            # Add description if available
            desc = f" - {node_.description}" if node_.description else ""
            lines.append(f"{name} ({status}){conn_str}{desc}")
        await ctx.output.print("\n".join(lines))
